use anyhow::Result;
use galactica_content::{Content, TextureHandle};
use std::{collections::HashMap, fs::File, io::Read, num::NonZeroU32};
use wgpu::BindGroupLayout;

use super::{rawtexture::RawTexture, Texture};

pub struct TextureArray {
	pub bind_group: wgpu::BindGroup,
	pub bind_group_layout: BindGroupLayout,
	starfield_handle: TextureHandle,
	textures: HashMap<TextureHandle, Texture>,
}

impl TextureArray {
	pub fn get_starfield_texture(&self) -> Texture {
		*self.textures.get(&self.starfield_handle).unwrap()
	}

	pub fn get_texture(&self, handle: TextureHandle) -> Texture {
		match self.textures.get(&handle) {
			Some(x) => *x,
			None => unreachable!("Tried to get a texture that doesn't exist"),
		}
	}

	pub fn new(device: &wgpu::Device, queue: &wgpu::Queue, ct: &Content) -> Result<Self> {
		// Load all textures
		let mut texture_data = Vec::new();
		let mut textures = HashMap::new();

		for t in &ct.textures {
			let mut f = File::open(&t.path)?;
			let mut bytes = Vec::new();
			f.read_to_end(&mut bytes)?;
			textures.insert(
				t.handle,
				Texture {
					index: texture_data.len() as u32,
					aspect: t.handle.aspect,
				},
			);

			texture_data.push(RawTexture::from_bytes(&device, &queue, &bytes, &t.name)?);
		}

		let sampler = device.create_sampler(&wgpu::SamplerDescriptor {
			address_mode_u: wgpu::AddressMode::ClampToEdge,
			address_mode_v: wgpu::AddressMode::ClampToEdge,
			address_mode_w: wgpu::AddressMode::ClampToEdge,
			mag_filter: wgpu::FilterMode::Linear,
			min_filter: wgpu::FilterMode::Nearest,
			mipmap_filter: wgpu::FilterMode::Nearest,
			..Default::default()
		});

		let bind_group_layout = device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
			label: Some("TextureArray bind group layout"),
			entries: &[
				// Texture data
				wgpu::BindGroupLayoutEntry {
					binding: 0,
					visibility: wgpu::ShaderStages::FRAGMENT,
					ty: wgpu::BindingType::Texture {
						multisampled: false,
						view_dimension: wgpu::TextureViewDimension::D2,
						sample_type: wgpu::TextureSampleType::Float { filterable: true },
					},
					count: NonZeroU32::new(texture_data.len() as u32),
				},
				// Texture sampler
				wgpu::BindGroupLayoutEntry {
					binding: 1,
					visibility: wgpu::ShaderStages::FRAGMENT,
					ty: wgpu::BindingType::Sampler(wgpu::SamplerBindingType::Filtering),
					count: NonZeroU32::new(1),
				},
			],
		});

		let views: Vec<&wgpu::TextureView> = texture_data.iter().map(|x| &x.view).collect();
		let bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
			label: Some("TextureArray bind group"),
			layout: &bind_group_layout,
			entries: &[
				wgpu::BindGroupEntry {
					binding: 0,
					// Array of all views
					resource: wgpu::BindingResource::TextureViewArray(&views),
				},
				wgpu::BindGroupEntry {
					binding: 1,
					resource: wgpu::BindingResource::SamplerArray(&[&sampler]),
				},
			],
		});

		return Ok(Self {
			bind_group,
			bind_group_layout,
			textures: textures,
			starfield_handle: ct.get_starfield_handle(),
		});
	}
}
